/*==============================================================================
*NOME:   T2AMNPAL.C
*
*  Projeto: Questo A do Teste 2 de Mtodos Numricos Para lgebra Linear
*
*  Verso:              V 1.00
*  Autor :   Juliana Carpes Imperial        - 9714169-4
*
*Descrio do mdulo de implementao
*     Objetivos:
*        - Implementar uma funo para calcular os autovalores de uma dada
*          matriz pelo mtodo de Jacobi;
*        - Implementar um programa principal que, dado um sistema, aplica
*          as funo acima citada e imprime resultados.
*=============================================================================*/

#include <stdio.h>
#include <math.h>

#define I_MAX_TAM 8           /* tamanho mximo da matriz */

/* strings para formatao de sada */
#define SZ_JACOBI1   "\nJACOBI UMA VEZ:\n"
#define SZ_JACOBI2   "\nJACOBI DUAS VEZES:\n"
#define SZ_AUTOVAL   "\nAUTOVALORES:\n"
#define SZ_MATRIZ_A  "A:\n"


/* definio das funes */

/*==============================================================================
*Nome: Jacobi
*Descrio da Funo: funo para calcular os auto-valores de uma matriz pelo
*                     mtodo de Jacobi.
*
*Parametros recebidos
*   A:  a matriz cujos auto-valores se deseja calcular;
*   iN: o tamanho do sistema.
*
*Parmetros retornados
*   Lambda: um vetor com a aproximao dos auto-valores dada pelo algoritmo;
*=============================================================================*/

void Jacobi( float A[ ][ I_MAX_TAM ] ,
             float Lambda[ ]         ,
             int   iN                 ) ;


/*==============================================================================
*Nome: ImprimirMatriz
*Descrio da Funo: funo para imprimir uma matriz.
*
*Parametros recebidos
*   M:  a matriz;
*   iN: o tamannho da matriz.
*=============================================================================*/

void ImprimirMatriz( float M[ ][ I_MAX_TAM ] ,
                     int   iN                 ) ;


/*==============================================================================
*Nome: ImprimirVetor
*Descrio da Funo: funo para imprimir um vetor.
*
*Parametros recebidos
*   V:  o vetor;
*   iN: o tamannho do vetor.
*=============================================================================*/

void ImprimirVetor( float V[ ] ,
                    int   iN    ) ;


/* implementao das funes */

/*==============================================================================
*Nome: Jacobi
*=============================================================================*/

void Jacobi( float A[ ][ I_MAX_TAM ] ,
             float Lambda[ ]         ,
             int   iN                 ) {

   float C[ I_MAX_TAM ][ I_MAX_TAM ] ;   /* matrz auxiliar */
   int   i                           ,   /* contador para as linhas de A */
         j                           ;   /* contador para as colunas de A */

/* para todas as linhas da matriz, faa */
   for ( j = 0 ; j < iN ; j++ ) {
      for ( i = 0 ; i < j ; i++ ) {

         float fD ,  /* o d do algoritmo */
               fE ,  /* o d com o ajuste do sinal */
               fC ,  /* o cosseno do ngulo */
               fS ;  /* o seno do ngulo */
         int   k  ,  /* contador para as linhas de C */
               l  ;  /* contador para as colunas de C */

         for ( k = 0 ; k < iN ; k++ ) {
            for ( l = 0 ; l < iN ; l++ ) {
               C[ k ][ l ] = A[ k ][ l ] ;
            } /* for */
         } /* for */

      /* fD = sqrt( ( Ai,i - Aj,j )^2 + 4 * Ai,j^2 ) */
         fD = pow( A[ i ][ i ] - A[ j ][ j ], 2. ) ;
         fD = sqrt( fD + 4. * A[ i ][ j ] * A[ i ][ j ] ) ;

      /* fE = sign( Ai,i - Aj,j ) * fD / 2 */
         fE = fD / 2. ;
         if ( ( A[ i ][ i ] - A[ j ][ j ] ) < 0. ) {
            fE *= -1. ;
         } /* if */

      /* fC = sqrt( ( 0.5 * ( 1 + | Ai,i - Aj,j | / fD ) ) */
         fC = sqrt( ( fD + fabs( A[ i ][ i ] - A[ j ][ j ] ) ) / ( 2. * fD ) ) ;

      /* fS = sign[ Ai,j( Ai,i - Aj,j ) ]
              sqrt( ( 0.5 * ( 1 - | Ai,i - Aj,j | / fD ) ) */
         fS = sqrt( ( fD - fabs( A[ i ][ i ] - A[ j ][ j ] ) ) / ( 2. * fD ) ) ;
         if ( ( A[ i ][ j ] * ( A[ i ][ i ] - A[ j ][ j ] ) ) < 0. ) {
            fS *= - 1. ;
         } /* if */

         C[ i ][ i ] = ( ( A[ i ][ i ] + A[ j ][ j ] ) / 2. ) + fE ;
         C[ j ][ j ] = ( ( A[ i ][ i ] + A[ j ][ j ] ) / 2. ) - fE ;
         C[ i ][ j ] = 0. ;
         C[ j ][ i ] = 0. ;

         for ( k = 0 ; k < iN ; k++ ) {
            if ( k != i
              && k != j ) {
               C[ i ][ k ] = fC * A[ k ][ i ] + fS * A[ k ][ j ] ;
               C[ k ][ i ] = fC * A[ k ][ i ] + fS * A[ k ][ j ] ;
               C[ k ][ j ] = - fS * A[ k ][ i ] + fC * A[ k ][ j ] ;
               C[ j ][ k ] = - fS * A[ k ][ i ] + fC * A[ k ][ j ] ;
            } /* if */
         } /* for */

         for ( k = 0 ; k < iN ; k++ ) {
            for ( l = 0 ; l < iN ; l++ ) {
               A[ k ][ l ] = C[ k ][ l ] ;
            } /* for */
         } /* for */

      } /* for */
   } /* for */

/* clculo dos valores aproximados dos autovalores de A */
   for ( i = 0 ; i < iN ; i++ ) {

   /* Li = Aii + S ( Aij^2 / ( Aii + Ajj ) ) */
      Lambda[ i ] = A[ i ][ i ] ;
      for ( j = 0 ; j < iN ; j++ ) {
         if ( i != j ) {
            Lambda[ i ] += pow( A[ i ][ j ], 2. )
                         / ( A[ i ][i ] + A[ j ][ j ] ) ;
         } /* if */
      } /* for */

   } /* for */

} /* Jacobi */


/*==============================================================================
*Nome: ImprimirMatriz
*=============================================================================*/

void ImprimirMatriz( float M[ ][ I_MAX_TAM ] ,
                     int   iN                 ) {

   int i ,  /* contador para linhas */
       j ;  /* contador para colunas */

/* imprime elementos de M */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {
         printf( "%25.21f", M[ i ][ j ] ) ;
      } /* for */
      printf( "\n" ) ;
   } /* for */

} /* ImprimirMatriz */


/*==============================================================================
*Nome: ImprimirVetor
*=============================================================================*/

void ImprimirVetor( float V[ ] ,
                    int   iN    ) {

   int i ;  /* contador */

/* imprime elementos de V */
   for ( i = 0 ; i < iN ; i++ ) {
      printf( "%.18f\n", V[ i ] ) ;
   } /* for */

   printf( "\n" ) ;

} /* ImprimirVetor */


int main( void ) {

   float A[ I_MAX_TAM ][ I_MAX_TAM ] , /* matriz A dada no teste, de n = 8 */
         Lambda[ I_MAX_TAM ]         ; /* auto-valores aproximados de A */
   int   i                           , /* contador de linhas para montar A */
         j                           ; /* contador de colunas para montar A */

/* monta a matriz A dada no teste */
   for ( i = 0 ; i < I_MAX_TAM ; i++ ) {

      for ( j = 0 ; j < I_MAX_TAM ; j++ ) {
         A[ i ][ j ] = 0. ;  /* inicializa matriz */
      } /* for */

      A[ i ][ i ] = 3. ;     /* seta diagonais */

   /* s seta as diagonais inferior e superior a principal se a linha no for a
      primeira nem a ltima */
      if ( i > 0 ) {
         A[ i ][ i - 1 ] = 1. ;
      } /* if */
      if ( i < I_MAX_TAM - 1 ) {
         A[ i ][ i + 1 ] = 1. ;
      } /* if */

   } /* for */

   A[ 0 ][ 0 ] = 2. ;                          /* seta A[ 0 ][ 0 ] */
   A[ I_MAX_TAM - 1 ][ I_MAX_TAM - 1 ] = 4. ;  /* seta A[ n ][ n ] */

/* imprime matriz A */
   printf( SZ_MATRIZ_A ) ;
   ImprimirMatriz( A, I_MAX_TAM ) ;

/* calcula autovalores por Jacobi pela primeira vez */
   printf( SZ_JACOBI1 ) ;
   Jacobi( A, Lambda, I_MAX_TAM ) ;
   ImprimirMatriz( A, I_MAX_TAM ) ;
   printf( SZ_AUTOVAL ) ;
   ImprimirVetor( Lambda, I_MAX_TAM ) ;

/* calcula autovalores por Jacobi pela primeira vez */
   printf( SZ_JACOBI2 ) ;
   Jacobi( A, Lambda, I_MAX_TAM ) ;
   ImprimirMatriz( A, I_MAX_TAM ) ;
   printf( SZ_AUTOVAL ) ;
   ImprimirVetor( Lambda, I_MAX_TAM ) ;

   return 0 ;

} /* main */

/* T2AMNPAL.C */
